
# THIS FILE IS GENERATED FROM SETUP.PY
short_version = '4.1.0'
version = '4.1.0'
full_version = '4.1.0'
git_revision = '7a74ef7f2a18b24d640ba8397822b8c2d2f8aa93'
release = True
if not release:
    version = full_version
